/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License, Version 1.0 only
 * (the "License").  You may not use this file except in compliance
 * with the License.
 *
 * You can obtain a copy of the license at http://smartos.org/CDDL
 *
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file.
 *
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 *
 * Copyright (c) 2016, Joyent, Inc. All rights reserved.
 *
 *
 * Unit tests for the firewall rule object
 */

var fwrule = require('../lib/index');
var util = require('util');



// Set this to any of the exports in this file to only run that test,
// plus setup and teardown
var runOne;



// --- Tests



exports['rule exports'] = function (t) {
    ['ACTIONS', 'DIRECTIONS', 'FIELDS', 'PROTOCOLS', 'TARGET_TYPES'].forEach(
        function (field) {
        t.ok(fwrule[field].length > 0, 'fwrule.' + field);
    });

    t.done();
};


exports['all target types'] = function (t) {
    var desc = 'all target types';
    var ips = ['192.168.1.1', '10.2.0.3'];
    var vms = ['9a343ca8-b42a-4a27-a9c5-800f57d1e8ed',
        '518908b6-8299-466d-8ea5-20a0ceff63ec'];
    var tags = ['tag1', 'tag2'];
    var subnets = ['192.168.2.0/24', '10.2.1.0/24'];
    var ruleTxt = util.format('FROM (ip %s OR vm %s OR tag %s OR subnet %s) ',
        ips[0], vms[0], tags[0], subnets[0])
        + util.format('TO (ip %s OR vm %s OR tag %s OR subnet %s)',
        ips[1], vms[1], tags[1], subnets[1])
        + ' ALLOW tcp port 80';

    var rule = fwrule.create({
        rule: ruleTxt,
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [ips[0]],
            subnets: [subnets[0]],
            vms: [vms[0]],
            tags: [tags[0]],
            wildcards: []
        },
        to: {
            ips: [ips[1]],
            subnets: [subnets[1]],
            vms: [vms[1]],
            tags: [tags[1]],
            wildcards: []
        },
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        ports: [ 80 ],
        action: 'allow',
        protocol: 'tcp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    var ser = {
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        global: true,
        rule: util.format('FROM (ip %s OR subnet %s OR tag %s OR vm %s) '
            + 'TO (ip %s OR subnet %s OR tag %s OR vm %s) ALLOW tcp PORT 80',
            ips[0], subnets[0], tags[0], vms[0],
            ips[1], subnets[1], tags[1], vms[1]),
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.serialize(), ser, 'rule.serialize()');
    t.deepEqual(rule.serialize(['enabled', 'version']),
        { enabled: ser.enabled, version: ser.version },
        'rule.serialize(): enabled, version');

    t.done();
};


exports['any'] = function (t) {
    var ip = '192.168.3.2';
    var vm = '8a343ca8-b42a-4a27-a9c5-800f57d1e8ed';
    var tag = 'tag3';
    var subnet = '192.168.0.0/16';

    var ruleTxt = util.format(
        'FROM (ip %s OR subnet %s OR tag %s OR vm %s) TO any ALLOW tcp PORT 80',
        ip, subnet, tag, vm);

    var rule = fwrule.create({
        rule: ruleTxt,
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [ip],
            subnets: [subnet],
            vms: [vm],
            tags: [tag],
            wildcards: []
        },
        to: {
            ips: [],
            subnets: [],
            vms: [],
            tags: [],
            wildcards: ['any']
        },
        enabled: true,
        ports: [ 80 ],
        action: 'allow',
        protocol: 'tcp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    t.deepEqual(rule.serialize(), {
        enabled: true,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.done();
};


exports['all vms'] = function (t) {
    var ip = '192.168.3.2';
    var owner = '50716241-ac8d-4e63-a9e4-77ff07cede61';

    var ruleTxt = util.format('FROM ip %s TO all vms ALLOW tcp PORT 80', ip);

    var rule = fwrule.create({
        rule: ruleTxt,
        enabled: true,
        owner_uuid: owner,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [ip],
            subnets: [],
            vms: [],
            tags: [],
            wildcards: []
        },
        to: {
            ips: [],
            subnets: [],
            vms: [],
            tags: [],
            wildcards: ['vmall']
        },
        enabled: true,
        owner_uuid: owner,
        ports: [ 80 ],
        action: 'allow',
        protocol: 'tcp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.deepEqual(rule.wildcards, raw.to.wildcards, 'rule.wildcards');
    t.ok(rule.allVMs, 'rule.allVMs');

    t.deepEqual(rule.serialize(), {
        enabled: true,
        // no global flag set because the rule has an owner_uuid
        owner_uuid: owner,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.done();
};


exports['tags'] = function (t) {
    var ruleTxt = 'FROM ip 1.2.3.4 TO tag some-tag ALLOW tcp PORT 80';
    var rule = new fwrule.create({
        rule: ruleTxt,
        version: fwrule.generateVersion()
    });

    var raw = {
        action: 'allow',
        enabled: false,
        from: {
            ips: [ '1.2.3.4' ],
            vms: [],
            subnets: [],
            tags: [],
            wildcards: []
        },
        protocol: 'tcp',
        ports: [ 80 ],
        to: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [ 'some-tag' ],
            wildcards: []
        },
        uuid: rule.uuid,
        version: rule.version
    };
    t.deepEqual(rule.raw(), raw, 'rule.raw()');

    t.deepEqual(rule.serialize(), {
        enabled: false,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');
    t.ok(!rule.allVMs, 'rule.allVMs');

    t.done();
};


exports['multiple ports and owner_uuid'] = function (t) {
    var inRule1 = {
        rule: 'FROM ip 10.88.88.1 TO tag tag2 ALLOW tcp '
            + '(PORT 1002 AND PORT 1052)',
        enabled: true,
        owner_uuid: '930896af-bf8c-48d4-885c-6573a94b1853',
        version: fwrule.generateVersion()
    };
    var inRule2 = {
        rule: 'FROM ip 10.88.88.1 TO tag tag2 ALLOW tcp '
            + 'PORTS 1002, 1052',
        enabled: true,
        owner_uuid: '930896af-bf8c-48d4-885c-6573a94b1853',
        version: fwrule.generateVersion()
    };


    var rule1 = fwrule.create(inRule1);
    var rule2 = fwrule.create(inRule2);
    var raw = {
        action: 'allow',
        enabled: inRule1.enabled,
        from: {
            ips: [ '10.88.88.1' ],
            vms: [],
            subnets: [],
            tags: [],
            wildcards: []
        },
        owner_uuid: inRule1.owner_uuid,
        protocol: 'tcp',
        ports: [ 1002, 1052 ],
        to: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [ 'tag2' ],
            wildcards: []
        },
        uuid: rule1.uuid,
        version: rule1.version
    };

    t.deepEqual(rule1.raw(), raw, 'rule1.raw()');
    t.deepEqual(rule1.ports, raw.ports, 'rule1.ports');
    t.deepEqual(rule1.protoTargets, raw.ports, 'rule1.protoTargets');

    t.deepEqual(rule1.serialize(), {
        enabled: true,
        owner_uuid: inRule1.owner_uuid,
        rule: inRule1.rule,
        uuid: rule1.uuid,
        version: rule1.version
    }, 'rule1.serialize()');

    raw.uuid = rule2.uuid;

    t.deepEqual(rule2.raw(), raw, 'rule2.raw()');
    t.deepEqual(rule2.ports, raw.ports, 'rule2.ports');
    t.deepEqual(rule2.protoTargets, raw.ports, 'rule2.protoTargets');

    t.deepEqual(rule2.serialize(), {
        enabled: true,
        owner_uuid: inRule2.owner_uuid,
        rule: inRule1.rule,
        uuid: rule2.uuid,
        version: rule2.version
    }, 'rule2.serialize()');

    t.done();
};


exports['icmp'] = function (t) {
    var vm = '8a343ca8-b42a-4a27-a9c5-800f57d1e8ed';

    var ruleTxt = util.format(
        'FROM any TO vm %s ALLOW icmp TYPE 8', vm);

    var rule = fwrule.create({
        rule: ruleTxt,
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [],
            subnets: [],
            vms: [],
            tags: [],
            wildcards: ['any']
        },
        to: {
            ips: [],
            subnets: [],
            vms: [vm],
            tags: [],
            wildcards: []
        },
        enabled: true,
        types: [ '8' ],
        action: 'allow',
        protocol: 'icmp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    t.deepEqual(rule.serialize(), {
        enabled: true,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.done();
};


exports['icmp with code'] = function (t) {
    var vm = '8a343ca8-b42a-4a27-a9c5-800f57d1e8ed';

    var ruleTxt = util.format(
        'FROM any TO vm %s ALLOW icmp TYPE 8 CODE 0', vm);

    var rule = fwrule.create({
        rule: ruleTxt,
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [],
            subnets: [],
            vms: [],
            tags: [],
            wildcards: ['any']
        },
        to: {
            ips: [],
            subnets: [],
            vms: [vm],
            tags: [],
            wildcards: []
        },
        enabled: true,
        types: [ '8:0' ],
        action: 'allow',
        protocol: 'icmp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    t.deepEqual(rule.serialize(), {
        enabled: true,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.done();
};


exports['icmp: multiple types'] = function (t) {
    var vm = '8a343ca8-b42a-4a27-a9c5-800f57d1e8ed';

    var ruleTxt = util.format(
        'FROM any TO vm %s ALLOW icmp (TYPE 8 CODE 0 AND TYPE 11 CODE 0 '
        + 'AND TYPE 30)', vm);

    var rule = fwrule.create({
        rule: ruleTxt,
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [],
            subnets: [],
            vms: [],
            tags: [],
            wildcards: ['any']
        },
        to: {
            ips: [],
            subnets: [],
            vms: [vm],
            tags: [],
            wildcards: []
        },
        enabled: true,
        types: [ '8:0', '11:0', '30' ],
        action: 'allow',
        protocol: 'icmp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    t.deepEqual(rule.types, raw.types, 'rule.types');
    t.deepEqual(rule.protoTargets, raw.types, 'rule.protoTargets');

    t.deepEqual(rule.serialize(), {
        enabled: true,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.done();
};


exports['sorting: icmp codes'] = function (t) {
    var vm = '8a343ca8-b42a-4a27-a9c5-800f57d1e8ed';

    var rule = fwrule.create({
        rule: util.format(
        'FROM any TO vm %s ALLOW icmp '
        + '(TYPE 8 CODE 0 AND TYPE 3 CODE 11 AND TYPE 40 AND TYPE 3 CODE 1 '
        + 'AND TYPE 30 AND TYPE 3 CODE 5)', vm),
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [],
            subnets: [],
            vms: [],
            tags: [],
            wildcards: ['any']
        },
        to: {
            ips: [],
            subnets: [],
            vms: [vm],
            tags: [],
            wildcards: []
        },
        enabled: true,
        types: [ '3:1', '3:5', '3:11', '8:0', '30', '40' ],
        action: 'allow',
        protocol: 'icmp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    t.deepEqual(rule.serialize(), {
        enabled: true,
        global: true,
        rule: util.format(
        'FROM any TO vm %s ALLOW icmp '
        + '(TYPE 3 CODE 1 AND TYPE 3 CODE 5 AND TYPE 3 CODE 11 '
        + 'AND TYPE 8 CODE 0 AND TYPE 30 AND TYPE 40)', vm),
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.done();
};

exports['sorting: icmp6 codes'] = function (t) {
    var vm = '8a343ca8-b42a-4a27-a9c5-800f57d1e8ed';

    var rule = fwrule.create({
        rule: util.format(
        'FROM any TO vm %s ALLOW icmp6 '
        + '(TYPE 8 CODE 0 AND TYPE 3 CODE 11 AND TYPE 40 AND TYPE 3 CODE 1 '
        + 'AND TYPE 30 AND TYPE 3 CODE 5)', vm),
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [],
            subnets: [],
            vms: [],
            tags: [],
            wildcards: ['any']
        },
        to: {
            ips: [],
            subnets: [],
            vms: [vm],
            tags: [],
            wildcards: []
        },
        enabled: true,
        types: [ '3:1', '3:5', '3:11', '8:0', '30', '40' ],
        action: 'allow',
        protocol: 'icmp6',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    t.deepEqual(rule.serialize(), {
        enabled: true,
        global: true,
        rule: util.format(
        'FROM any TO vm %s ALLOW icmp6 '
        + '(TYPE 3 CODE 1 AND TYPE 3 CODE 5 AND TYPE 3 CODE 11 '
        + 'AND TYPE 8 CODE 0 AND TYPE 30 AND TYPE 40)', vm),
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.done();
};

exports['sorting: ports'] = function (t) {
    var inRule = {
        rule: 'FROM ip 10.88.88.1 TO tag tag2 ALLOW tcp '
            + '(PORT 1002 AND PORT 10 AND PORT 1052 AND PORT 80 AND PORT 30245 '
            + 'AND PORT 6)',
        enabled: true,
        version: fwrule.generateVersion()
    };

    var rule = fwrule.create(inRule);
    var raw = {
        action: 'allow',
        enabled: inRule.enabled,
        from: {
            ips: [ '10.88.88.1' ],
            vms: [],
            subnets: [],
            tags: [],
            wildcards: []
        },
        protocol: 'tcp',
        ports: [ 6, 10, 80, 1002, 1052, 30245 ],
        to: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [ 'tag2' ],
            wildcards: []
        },
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');

    t.deepEqual(rule.serialize(), {
        enabled: true,
        global: true,
        rule: 'FROM ip 10.88.88.1 TO tag tag2 ALLOW tcp '
            + '(PORT 6 AND PORT 10 AND PORT 80 AND PORT 1002 AND PORT 1052 '
            + 'AND PORT 30245)',
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.done();
};

exports['sorting: port ranges'] = function (t) {
    var inRule = {
        rule: 'FROM ip 10.88.88.1 TO tag tag2 ALLOW tcp '
            + 'PORTS 1002, 20-40, 10, 1052, 80, 30245, 6 - 11',
        enabled: true,
        version: fwrule.generateVersion()
    };

    var rule = fwrule.create(inRule);
    var raw = {
        action: 'allow',
        enabled: inRule.enabled,
        from: {
            ips: [ '10.88.88.1' ],
            vms: [],
            subnets: [],
            tags: [],
            wildcards: []
        },
        protocol: 'tcp',
        ports: [ { start: 6, end: 11 }, 10, { start: 20, end: 40 },
            80, 1002, 1052, 30245 ],
        to: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [ 'tag2' ],
            wildcards: []
        },
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.strictEqual(rule.raw().ports[0].start, raw.ports[0].start,
        'Both starts are Numbers');
    t.strictEqual(rule.raw().ports[0].end, raw.ports[0].end,
        'Both ends are Numbers');

    t.deepEqual(rule.serialize(), {
        enabled: true,
        global: true,
        rule: 'FROM ip 10.88.88.1 TO tag tag2 ALLOW tcp '
            + 'PORTS 6 - 11, 10, 20 - 40, 80, 1002, 1052, 30245',
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.done();
};

exports['single port range'] = function (t) {
    var inRule = {
        rule: 'FROM ip 10.88.88.1 TO tag tag2 ALLOW tcp '
            + 'PORTS 50-50',
        enabled: true,
        version: fwrule.generateVersion()
    };

    var rule = fwrule.create(inRule);
    var raw = {
        action: 'allow',
        enabled: inRule.enabled,
        from: {
            ips: [ '10.88.88.1' ],
            vms: [],
            subnets: [],
            tags: [],
            wildcards: []
        },
        protocol: 'tcp',
        ports: [ { start: 50, end: 50 } ],
        to: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [ 'tag2' ],
            wildcards: []
        },
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');

    t.deepEqual(rule.serialize(), {
        enabled: true,
        global: true,
        rule: 'FROM ip 10.88.88.1 TO tag tag2 ALLOW tcp '
            + 'PORTS 50 - 50',
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.done();
};

exports['port ALL'] = function (t) {
    var normalText = 'FROM ip 10.88.88.1 TO tag tag2 ALLOW tcp PORT all';
    var parenText = 'FROM ip 10.88.88.1 TO tag tag2 ALLOW tcp ( PORT all )';
    var ruleTexts = [ normalText, parenText ];

    ruleTexts.forEach(function (ruleText) {
        var inRule = {
            rule: ruleText,
            enabled: true,
            version: fwrule.generateVersion()
        };

        var rule = fwrule.create(inRule);
        var raw = {
            action: 'allow',
            enabled: inRule.enabled,
            from: {
                ips: [ '10.88.88.1' ],
                vms: [],
                subnets: [],
                tags: [],
                wildcards: []
            },
            protocol: 'tcp',
            ports: [ 'all' ],
            to: {
                ips: [],
                vms: [],
                subnets: [],
                tags: [ 'tag2' ],
                wildcards: []
            },
            uuid: rule.uuid,
            version: rule.version
        };

        t.deepEqual(rule.raw(), raw, 'rule.raw()');
        t.deepEqual(rule.ports, raw.ports, 'rule.ports');
        t.deepEqual(rule.protoTargets, raw.ports, 'rule.protoTargets');

        t.deepEqual(rule.serialize(), {
            enabled: true,
            global: true,
            rule: normalText,
            uuid: rule.uuid,
            version: rule.version
        }, 'rule.serialize()');
    });

    t.done();
};


exports['tags: equal'] = function (t) {
    var ruleTxt =
        'FROM ip 1.2.3.4 TO tag some-tag = value ALLOW tcp PORT 80';
    var rule = new fwrule.create({
        rule: ruleTxt,
        version: fwrule.generateVersion()
    });

    var raw = {
        action: 'allow',
        enabled: false,
        from: {
            ips: [ '1.2.3.4' ],
            vms: [],
            subnets: [],
            tags: [],
            wildcards: []
        },
        protocol: 'tcp',
        ports: [ 80 ],
        to: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [ [ 'some-tag', 'value' ] ],
            wildcards: []
        },
        uuid: rule.uuid,
        version: rule.version
    };
    t.deepEqual(rule.raw(), raw, 'rule.raw()');

    t.deepEqual(rule.serialize(), {
        enabled: false,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.ok(!rule.allVMs, 'rule.allVMs');
    t.deepEqual(rule.tags, raw.to.tags, 'rule.tags');

    t.done();
};


exports['multiple tags: equal'] = function (t) {
    var ruleTxt = 'FROM ip 1.2.3.4 TO '
        + '(tag some-tag = value OR tag some-tag = value2) ALLOW tcp PORT 80';
    var rule = new fwrule.create({
        rule: ruleTxt,
        version: fwrule.generateVersion()
    });

    var raw = {
        action: 'allow',
        enabled: false,
        from: {
            ips: [ '1.2.3.4' ],
            vms: [],
            subnets: [],
            tags: [],
            wildcards: []
        },
        protocol: 'tcp',
        ports: [ 80 ],
        to: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [
                [ 'some-tag', 'value' ],
                [ 'some-tag', 'value2' ]
            ],
            wildcards: []
        },
        uuid: rule.uuid,
        version: rule.version
    };
    t.deepEqual(rule.raw(), raw, 'rule.raw()');

    t.deepEqual(rule.serialize(), {
        enabled: false,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.ok(!rule.allVMs, 'rule.allVMs');
    t.deepEqual(rule.tags, raw.to.tags, 'rule.tags');

    t.done();
};


exports['multiple tags: multiple values'] = function (t) {
    var rule = new fwrule.create({
        rule: 'FROM (tag some-tag OR tag some-tag = value0) TO '
            + '(tag some-tag = value OR tag some-tag = value2) '
            + 'ALLOW tcp PORT 80',
        version: fwrule.generateVersion()
    });

    var raw = {
        action: 'allow',
        enabled: false,
        from: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [ 'some-tag' ],
            wildcards: []
        },
        protocol: 'tcp',
        ports: [ 80 ],
        to: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [
                [ 'some-tag', 'value' ],
                [ 'some-tag', 'value2' ]
            ],
            wildcards: []
        },
        uuid: rule.uuid,
        version: rule.version
    };
    t.deepEqual(rule.raw(), raw, 'rule.raw()');

    t.deepEqual(rule.serialize(), {
        enabled: false,
        global: true,
        // 'some-tag = value0' is a subset of 'tag some-tag', so it is not
        // included in the rule text
        rule: 'FROM tag some-tag TO '
            + '(tag some-tag = value OR tag some-tag = value2) '
            + 'ALLOW tcp PORT 80',
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.ok(!rule.allVMs, 'rule.allVMs');
    t.deepEqual(rule.tags, raw.from.tags, 'rule.tags');

    t.done();
};


exports['multiple tags: multiple quoted values'] = function (t) {
    var owner = 'ace1da4b-9ab2-4991-8298-700bec1b70ed';
    var rule = new fwrule.create({
        owner_uuid: owner,
        rule: 'FROM '
            + '(tag "김치" = "백김치" '
            + 'OR tag "김치" = "白김치") TO '
            + '(tag "some tag" = value '
            + 'OR tag some-tag = "another value") '
            + 'ALLOW tcp PORT 80',
        version: fwrule.generateVersion()
    });

    var raw = {
        action: 'allow',
        enabled: false,
        from: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [
                [ '김치', '白김치' ],
                [ '김치', '백김치' ]
            ],
            wildcards: []
        },
        owner_uuid: owner,
        protocol: 'tcp',
        ports: [ 80 ],
        to: {
            ips: [],
            vms: [],
            subnets: [],
            tags: [
                [ 'some tag', 'value' ],
                [ 'some-tag', 'another value' ]
            ],
            wildcards: []
        },
        uuid: rule.uuid,
        version: rule.version
    };
    t.deepEqual(rule.raw(), raw, 'rule.raw()');

    t.deepEqual(rule.serialize(), {
        enabled: false,
        owner_uuid: owner,
        rule: 'FROM (tag "김치" = "白김치" '
            + 'OR tag "김치" = "백김치") TO '
            + '(tag "some tag" = value OR tag some-tag = "another value") '
            + 'ALLOW tcp PORT 80',
        uuid: rule.uuid,
        version: rule.version
    }, 'rule.serialize()');

    t.ok(!rule.allVMs, 'rule.allVMs');
    t.deepEqual(rule.tags, [
                [ 'some tag', 'value' ],
                [ 'some-tag', 'another value' ],
                [ '김치', '白김치' ],
                [ '김치', '백김치' ]
        ], 'rule.tags');

    t.done();
};


exports['IPv6 sources'] = function (t) {
    var desc = 'IPv6 sources';
    var vm = '9a343ca8-b42a-4a27-a9c5-800f57d1e8ed';
    var ips = ['fd00::2', 'fe80::8:20ff:fe40:65e4'];
    var ruleTxt = util.format('FROM (ip %s OR ip %s) ', ips[0], ips[1])
        + util.format('TO vm %s ALLOW tcp PORT 80', vm);

    var rule = fwrule.create({
        rule: ruleTxt,
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: ips,
            subnets: [],
            vms: [],
            tags: [],
            wildcards: []
        },
        to: {
            ips: [],
            subnets: [],
            vms: [vm],
            tags: [],
            wildcards: []
        },
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        ports: [ 80 ],
        action: 'allow',
        protocol: 'tcp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    var ser = {
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.serialize(), ser, 'rule.serialize()');
    t.deepEqual(rule.serialize(['enabled', 'version']),
        { enabled: ser.enabled, version: ser.version },
        'rule.serialize(): enabled, version');

    t.done();
};

exports['IPv6 subnet sources'] = function (t) {
    var desc = 'IPv6 subnet sources';
    var vm = '9a343ca8-b42a-4a27-a9c5-800f57d1e8ed';
    var cidr = 'fd00::/64';
    var ruleTxt = util.format('FROM subnet %s ', cidr)
        + util.format('TO vm %s ALLOW tcp PORT 80', vm);

    var rule = fwrule.create({
        rule: ruleTxt,
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [],
            subnets: [cidr],
            vms: [],
            tags: [],
            wildcards: []
        },
        to: {
            ips: [],
            subnets: [],
            vms: [vm],
            tags: [],
            wildcards: []
        },
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        ports: [ 80 ],
        action: 'allow',
        protocol: 'tcp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    var ser = {
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.serialize(), ser, 'rule.serialize()');
    t.deepEqual(rule.serialize(['enabled', 'version']),
        { enabled: ser.enabled, version: ser.version },
        'rule.serialize(): enabled, version');

    t.done();
};

exports['IPv6 destinations'] = function (t) {
    var desc = 'IPv6 destinations';
    var vm = '9a343ca8-b42a-4a27-a9c5-800f57d1e8ed';
    var ips = ['fd00::1', 'fd00::2'];
    var ruleTxt = util.format('FROM vm %s ', vm)
        + util.format('TO (ip %s OR ip %s) ALLOW tcp PORT 80', ips[0], ips[1]);

    var rule = fwrule.create({
        rule: ruleTxt,
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [],
            subnets: [],
            vms: [vm],
            tags: [],
            wildcards: []
        },
        to: {
            ips: ips,
            subnets: [],
            vms: [],
            tags: [],
            wildcards: []
        },
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        ports: [ 80 ],
        action: 'allow',
        protocol: 'tcp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    var ser = {
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.serialize(), ser, 'rule.serialize()');
    t.deepEqual(rule.serialize(['enabled', 'version']),
        { enabled: ser.enabled, version: ser.version },
        'rule.serialize(): enabled, version');

    t.done();
};

exports['IPv6 subnet destinations'] = function (t) {
    var desc = 'IPv6 subnet destinations';
    var vm = '9a343ca8-b42a-4a27-a9c5-800f57d1e8ed';
    var cidr = 'fd00::/64';
    var ruleTxt = util.format('FROM vm %s ', vm)
        + util.format('TO subnet %s BLOCK tcp PORT 80', cidr);

    var rule = fwrule.create({
        rule: ruleTxt,
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: [],
            subnets: [],
            vms: [vm],
            tags: [],
            wildcards: []
        },
        to: {
            ips: [],
            subnets: [cidr],
            vms: [],
            tags: [],
            wildcards: []
        },
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        ports: [ 80 ],
        action: 'block',
        protocol: 'tcp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    var ser = {
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.serialize(), ser, 'rule.serialize()');
    t.deepEqual(rule.serialize(['enabled', 'version']),
        { enabled: ser.enabled, version: ser.version },
        'rule.serialize(): enabled, version');

    t.done();
};

exports['Mixed IPv4 and IPv6'] = function (t) {
    var desc = 'Mixed IPv4 and IPv6';
    var vm1 = '9a343ca8-b42a-4a27-a9c5-800f57d1e8ed';
    var vm2 = '518908b6-8299-466d-8ea5-20a0ceff63ec';
    var ips = ['10.10.10.5', 'fd00::1'];
    var ruleTxt =
        util.format('FROM (ip %s OR ip %s OR vm %s) ', ips[0], ips[1], vm1)
        + util.format('TO vm %s ALLOW tcp PORT 80', vm2);

    var rule = fwrule.create({
        rule: ruleTxt,
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        version: fwrule.generateVersion()
    });

    var raw = {
        from: {
            ips: ips,
            subnets: [],
            vms: [vm1],
            tags: [],
            wildcards: []
        },
        to: {
            ips: [],
            subnets: [],
            vms: [vm2],
            tags: [],
            wildcards: []
        },
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        ports: [ 80 ],
        action: 'allow',
        protocol: 'tcp',
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.raw(), raw, 'rule.raw()');
    t.deepEqual(rule.from, raw.from, 'rule.from');
    t.deepEqual(rule.to, raw.to, 'rule.to');
    t.ok(!rule.allVMs, 'rule.allVMs');

    var ser = {
        created_by: 'fwadm',
        description: desc,
        enabled: true,
        global: true,
        rule: ruleTxt,
        uuid: rule.uuid,
        version: rule.version
    };

    t.deepEqual(rule.serialize(), ser, 'rule.serialize()');
    t.deepEqual(rule.serialize(['enabled', 'version']),
        { enabled: ser.enabled, version: ser.version },
        'rule.serialize(): enabled, version');

    t.done();
};



// Use to run only one test in this file:
if (runOne) {
    module.exports = {
        oneTest: runOne
    };
}
