# Overview

The vmload module converts system data into VM objects. Merging data from:

 * /etc/zones/<zonename>.xml
 * sysinfo
 * zfs
 * zoneadm
 * /zones/<zonename>/config/{tags,metadata,routes}.json
 * /zones/<zonename>/lastexited

into a single JSON representation of the zone. It utilizes the data in
../proptable.js to control how values are loaded.

This module is currently intended to be used only by VM.js / vmadm and this
documentation exists for internal use. The interface should be considered
unstable and may be changed (along with the VM.js code) at any time.


# IMPORTANT Warning

When using getZoneData() you can restrict the values to certain fields by
passing a 'fields' option. You should only do this when you understand the
consequences. Specifically, if you create a 'cache' with a limited set of
fields, and then use that same cache for another getVmobj call which requires
different fields to satisfy, you will not have all the possible data in the
output object.

The way to avoid this is to always load the full cache without passing a
'fields' option to getZonedata() when you intend to reuse the cache unless you
understand exactly what you're doing.

If you call getVmobj() or getVmobjs() without a 'cache' option you will never
run into this problem as data is (re)loaded for each request. This obviously
has a performance penalty but will be safe.


# Functions

## getVmobj(uuid, options, callback)

This builds a JSON object for a single VM with UUID matching the uuid parameter
on the system and calls:

  callback(err, <object>);

once the object is complete. The 'options' supported are:

  cache: a cache object output from getZoneData() (optional)
  fields: an array of fields to include in the output object (optional)
  log: a bunyan logger (required)

when 'fields' is not set, the object passed to the callback will contain all
available fields.

When passing a 'cache' option, see "IMPORTANT Warning" section above if you've
passed a 'fields' option to getZoneData.


## getVmobjs(filter, options, callback)

This builds JSON objects representing zero or more VMs on the system for which
the filter matches. The 'options' supported are:

  cache: a cache object output from getZoneData() (optional)
  fields: an array of fields to include in the output object (optional)
  log: a bunyan logger (required)

The filter is a function called like:

  filter(vmobj,cb);

and the filter function must call cb(<true|false>) where true includes the
object in the result and false does not. After going through all the system's
VMs and applying the filter getVmobjs calls:

  callback(err, <array of VM objects>);

with an array of all the VMs that matched.

When passing a 'cache' option, see "IMPORTANT Warning" section above if you've
passed a 'fields' option to getZoneData.

## getZoneData(uuid, options, callback)

This is used to build a 'cache' object that can be passed into getVmobj and/or
getVmobjs. The idea here is that when you might want to run multiple lookups,
you can do:

 getZoneData(null, options, function (err, cache) {
     ...
     options.cache = cache;
     getVmobj(uuid, options, function (err, vmobj) {
         ...
         <do something with vmobj>
         ...
     });
     ...
 });

and if getVmobj is called multiple times with the same cache, the data will not
need to be loaded again. This is especially useful if you're able to watch for
zone changes and reload the appropriate parts of the cache when those components
change.

See the "IMPORTANT Warning" section above for cautionary notes on using this when
combined with options.fields.

# Examples

## Calling getVmobjs to get an object containing all VMs

(see dump-vmobjs.js)

## Calling getVmobj to get a single VM

Assuming you have a VM with UUID 8be21e2a-ce25-4eb2-b796-b36b274b5a25:

    var bunyan = require('bunyan');
    var getVmobj = require('vmload').getVmobj;

    var log = bunyan.createLogger({
        level: 'debug',
        name: 'dump-vmobjs',
        streams: [
            {
                stream: process.stderr,
                level: 'warn'
            }
        ],
        serializers: bunyan.stdSerializers
    });

    getVmobj('8be21e2a-ce25-4eb2-b796-b36b274b5a25', {log: log},
        function (err, obj) {
            if (err) {
                throw (err);
            }
            console.log(JSON.stringify(obj, null, 2));
        }
    );

Output looks like:

    {
      "zonename": "8be21e2a-ce25-4eb2-b796-b36b274b5a25",
      "autoboot": true,
      "brand": "joyent-minimal",
      "limit_priv": "default",
      ...
    }

## Calling getVmobj to get only a subset of properties:

Again assuming you have a VM with UUID 8be21e2a-ce25-4eb2-b796-b36b274b5a25:

    var bunyan = require('bunyan');
    var getVmobj = require('vmload').getVmobj;

    var log = bunyan.createLogger({
        level: 'debug',
        name: 'dump-vmobjs',
        streams: [
            {
                stream: process.stderr,
                level: 'warn'
            }
        ],
        serializers: bunyan.stdSerializers
    });

    getVmobj('8be21e2a-ce25-4eb2-b796-b36b274b5a25',
        {
            log: log,
            fields: ['zonename', 'state']
        },
        function (err, obj) {
            if (err) {
                throw (err);
            }
            console.log(JSON.stringify(obj, null, 2));
        }
    );

Output will look like:

    {
      "zonename": "8be21e2a-ce25-4eb2-b796-b36b274b5a25",
      "state": "running"
    }

and internally we'll not have had to load any ZFS data or JSON data so this
will be faster than loading the whole object if we only need these fields.
